/*--------------------------------------------------------------------------*/
/*  Copyright (C) Herbert Oppmann 2003. All Rights Reserved.                */
/*--------------------------------------------------------------------------*/
/*  Project:     MyBCP                                                      */
/*  Filename:    LZHPack.c                                                  */
/*                                                                          */
/*  Description:                                                            */
/*                                                                          */
/*  LZH compression implementation                                          */
/*--------------------------------------------------------------------------*/

/*--< define, include >-----------------------------------------------------*/

#include "LZHapi.h"
#include "LZHint.h"
#include <mem.h>	// memmove

#define PERC_FLAG 0x8000U
#define PERCOLATE  1
#define NIL        0
#define MAX_HASH_VAL (3 * DDICSIZ + (DDICSIZ / 512 + 1) * UCHAR_MAX)
#define HASH(p, c) ((p) + ((c) << (MAXDICBIT - 9)) + DDICSIZ * 2)

/*--< typedef >-------------------------------------------------------------*/

/* Lempel-Ziv 77 sliding window */
typedef short node;

/*--< variables >-----------------------------------------------------------*/

/* Lempel-Ziv 77 sliding window */
static uchar *text, *childcount;
static node pos, matchpos, avail,
	*position, *parent, *prev, *next = NULL;
static int remainder, matchlen;
#if MAXMATCH <= (UCHAR_MAX + 1)
	static uchar *level;
#else
	static ushort *level;
#endif
/* static Huffman encoding */
static int    n, heapsize;
static short  heap[NC + 1];
static ushort *freq, *sortptr, len_cnt[17];
static uchar  *len;
static ushort c_freq[2 * NC - 1], c_code[NC],
              p_freq[2 * NP - 1], pt_code[NPT],
              t_freq[2 * NT - 1];
static uint   bufsiz = 0;
static uchar *buf;
static uint output_pos, output_mask;

/*--< function declarations >-----------------------------------------------*/

/* Lempel-Ziv 77 sliding window */
static void allocate_memory(void);
static void init_slide(void);
static void get_next_match(void);
static void insert_node(void);
static void delete_node(void);
static node child(node q, uchar c);
static void makechild(node q, uchar c, node r);
static void split(node old);
/* static Huffman encoding */
static void huf_encode_start(void);
static void huf_encode_end(void);
static void output(uint c, uint p);
static void send_block(void);
static int make_tree(int nparm, ushort freqparm[],
				uchar lenparm[], ushort codeparm[]);
static void count_t_freq(void);
static void write_pt_len(int n, int nbit, int i_special);
static void write_c_len(void);
static void encode_c(int c);
static void encode_p(uint p);
static void downheap(int i);
static void make_len(int root);
static void make_code(int n, uchar len[], ushort code[]);
static void count_len(int i);
/* variable number of output bits */
static void init_putbits(void);
static void putbits(int n, uint x);

/*--< functions >-----------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* Lempel-Ziv 77 sliding window                                             */
/*--------------------------------------------------------------------------*/


void encode(void)
{
	int lastmatchlen;
	node lastmatchpos;

	allocate_memory();
    init_slide();
    huf_encode_start();
	remainder = read_uncompressed(&text[DDICSIZ], DDICSIZ + MAXMATCH);
	matchlen = 0;
	pos = DDICSIZ;
    insert_node();
	if (matchlen > remainder)
    	matchlen = remainder;
	while (remainder > 0) {
		lastmatchlen = matchlen;
        lastmatchpos = matchpos;
		get_next_match();
		if (matchlen > remainder)
        	matchlen = remainder;
		if (matchlen > lastmatchlen || lastmatchlen < THRESHOLD)
			output(text[pos - 1], 0);
		else {
			output(lastmatchlen + (UCHAR_MAX + 1 - THRESHOLD),
				   (pos - lastmatchpos - 2) & (DDICSIZ - 1));
			while (--lastmatchlen > 0)
            	get_next_match();
			if (matchlen > remainder)
            	matchlen = remainder;
		}
	}
	huf_encode_end();
}


/* allocate buffers */
static void allocate_memory(void)
{
	if (next != NULL)
    	return;
    text       = malloc(DDICSIZ * 2 + MAXMATCH);
	level      = malloc((DDICSIZ + UCHAR_MAX + 1) * sizeof(*level));
	childcount = malloc((DDICSIZ + UCHAR_MAX + 1) * sizeof(*childcount));
	#if PERCOLATE
	  position = malloc((DDICSIZ + UCHAR_MAX + 1) * sizeof(*position));
	#else
	  position = malloc(DDICSIZ * sizeof(*position));
	#endif
	parent     = malloc(DDICSIZ * 2 * sizeof(*parent));
	prev       = malloc(DDICSIZ * 2 * sizeof(*prev));
	next       = malloc((MAX_HASH_VAL + 1) * sizeof(*next));
	if (next == NULL)
    	error("Out of memory.");
}


static void init_slide(void)
{
	node i;

	for (i = DDICSIZ; i <= DDICSIZ + UCHAR_MAX; i++) {
		level[i] = 1;
		#if PERCOLATE
			position[i] = NIL;  /* sentinel */
		#endif
	}
	for (i = DDICSIZ; i < DDICSIZ * 2; i++)
    	parent[i] = NIL;
	avail = 1;
	for (i = 1; i < DDICSIZ - 1; i++)
    	next[i] = i + 1;
	next[DDICSIZ - 1] = NIL;
	for (i = DDICSIZ * 2; i <= MAX_HASH_VAL; i++)
    	next[i] = NIL;
}


static void get_next_match(void)
{
	int n;

	remainder--;
	if (++pos == DDICSIZ * 2) {
		memmove(&text[0], &text[DDICSIZ], DDICSIZ + MAXMATCH);
		n = read_uncompressed(&text[DDICSIZ + MAXMATCH], DDICSIZ);
		remainder += n;
        pos = DDICSIZ;
	}
	delete_node();
    insert_node();
}


static void insert_node(void)
{
	node q, r, j, t;
	uchar c, *t1, *t2;

	if (matchlen >= 4) {
		matchlen--;
		r = (matchpos + 1) | DDICSIZ;
		while ((q = parent[r]) == NIL)
        	r = next[r];
		while (level[q] >= matchlen) {
			r = q;
            q = parent[q];
		}
		#if PERCOLATE
			t = q;
			while (position[t] < 0) {
				position[t] = pos;
                t = parent[t];
			}
			if (t < DDICSIZ)
            	position[t] = pos | PERC_FLAG;
		#else
			t = q;
			while (t < DDICSIZ) {
				position[t] = pos;
                t = parent[t];
			}
		#endif
	} else {
		q = text[pos] + DDICSIZ;
        c = text[pos + 1];
		if ((r = child(q, c)) == NIL) {
			makechild(q, c, pos);
            matchlen = 1;
			return;
		}
		matchlen = 2;
	}
	for ( ; ; ) {
		if (r >= DDICSIZ) {
			j = MAXMATCH;
            matchpos = r;
		} else {
			j = level[r];
			matchpos = position[r] & ~PERC_FLAG;
		}
		if (matchpos >= pos)
        	matchpos -= DDICSIZ;
		t1 = &text[pos + matchlen];
        t2 = &text[matchpos + matchlen];
		while (matchlen < j) {
			if (*t1 != *t2) {
            	split(r);
                return;
            }
			matchlen++;
            t1++;
            t2++;
		}
		if (matchlen >= MAXMATCH)
        	break;
		position[r] = pos;
		q = r;
		if ((r = child(q, *t1)) == NIL) {
			makechild(q, *t1, pos);
            return;
		}
		matchlen++;
	}
	t = prev[r];
    prev[pos] = t;
    next[t] = pos;
	t = next[r];
    next[pos] = t;
    prev[t] = pos;
	parent[pos] = q;
    parent[r] = NIL;
	next[r] = pos;  /* special use of next[] */
}


static void delete_node(void)
{
	#if PERCOLATE
		node q, r, s, t, u;
	#else
		node r, s, t, u;
	#endif

	if (parent[pos] == NIL)
    	return;
	r = prev[pos];
    s = next[pos];
	next[r] = s;
    prev[s] = r;
	r = parent[pos];
    parent[pos] = NIL;
	if (r >= DDICSIZ || --childcount[r] > 1)
    	return;
	#if PERCOLATE
		t = position[r] & ~PERC_FLAG;
	#else
		t = position[r];
	#endif
	if (t >= pos)
    	t -= DDICSIZ;
	#if PERCOLATE
		s = t;
        q = parent[r];
		while ((u = position[q]) & PERC_FLAG) {
			u &= ~PERC_FLAG;
            if (u >= pos)
            	u -= DDICSIZ;
			if (u > s)
            	s = u;
			position[q] = (s | DDICSIZ);
            q = parent[q];
		}
		if (q < DDICSIZ) {
			if (u >= pos)
            	u -= DDICSIZ;
			if (u > s)
            	s = u;
			position[q] = s | DDICSIZ | PERC_FLAG;
		}
	#endif
	s = child(r, text[t + level[r]]);
	t = prev[s];
    u = next[s];
	next[t] = u;
    prev[u] = t;
	t = prev[r];
    next[t] = s;
    prev[s] = t;
	t = next[r];
    prev[t] = s;
    next[s] = t;
	parent[s] = parent[r];
    parent[r] = NIL;
	next[r] = avail;
    avail = r;
}


/* q's child for character c (NIL if not found) */
static node child(node q, uchar c)
{
	node r;

	r = next[HASH(q, c)];
	parent[NIL] = q;  /* sentinel */
	while (parent[r] != q)
    	r = next[r];
	return r;
}


/* Let r be q's child for character c. */
static void makechild(node q, uchar c, node r)
{
	node h, t;

	h = HASH(q, c);
	t = next[h];
    next[h] = r;
    next[r] = t;
	prev[t] = r;
    prev[r] = h;
	parent[r] = q;
    childcount[q]++;
}


static void split(node old)
{
	node new, t;

	new = avail;
    avail = next[new];
    childcount[new] = 0;
	t = prev[old];
    prev[new] = t;
    next[t] = new;
	t = next[old];
    next[new] = t;
    prev[t] = new;
	parent[new] = parent[old];
	level[new] = matchlen;
	position[new] = pos;
	makechild(new, text[matchpos + matchlen], old);
	makechild(new, text[pos + matchlen], pos);
}


/*--------------------------------------------------------------------------*/
/* static Huffman encoding                                                  */
/*--------------------------------------------------------------------------*/


static void huf_encode_start(void)
{
	int i;

	if (bufsiz == 0) {
		bufsiz = 16 * 1024U;
		while ((buf = malloc(bufsiz)) == NULL) {
			bufsiz = (bufsiz / 10U) * 9U;
			if (bufsiz < 4 * 1024U)
            	error("Out of memory.");
		}
	}
	buf[0] = 0;
	for (i = 0; i < NC; i++)
    	c_freq[i] = 0;
	for (i = 0; i < NP; i++)
    	p_freq[i] = 0;
	output_pos = output_mask = 0;
	init_putbits();
}


static void huf_encode_end(void)
{
	send_block();
	putbits(CHAR_BIT - 1, 0);  /* flush remaining bits */
}


static void output(uint c, uint p)
{
	static uint cpos;

	if ((output_mask >>= 1) == 0) {
		output_mask = 1U << (CHAR_BIT - 1);
		if (output_pos >= bufsiz - 3 * CHAR_BIT) {
			send_block();
			output_pos = 0;
		}
		cpos = output_pos++;
        buf[cpos] = 0;
	}
	buf[output_pos++] = (uchar) c;
    c_freq[c]++;
	if (c >= (1U << CHAR_BIT)) {
		buf[cpos] |= output_mask;
		buf[output_pos++] = (uchar)(p >> CHAR_BIT);
		buf[output_pos++] = (uchar) p;
		c = 0;
        while (p) {
        	p >>= 1;
            c++;
        }
		p_freq[c]++;
	}
}


static void send_block(void)
{
	uint i, k, flags, root, pos, size;

	root = make_tree(NC, c_freq, c_len, c_code);
	size = c_freq[root];
    putbits(16, size);
	if (root >= NC) {
		count_t_freq();
		root = make_tree(NT, t_freq, pt_len, pt_code);
		if (root >= NT) {
			write_pt_len(NT, TBIT, 3);
		} else {
			putbits(TBIT, 0);
            putbits(TBIT, root);
		}
		write_c_len();
	} else {
        putbits(TBIT, 0);
        putbits(TBIT, 0);
		putbits(CBIT, 0);
        putbits(CBIT, root);
	}
	root = make_tree(NP, p_freq, pt_len, pt_code);
	if (root >= NP) {
		write_pt_len(NP, PBIT, -1);
	} else {
		putbits(PBIT, 0);
        putbits(PBIT, root);
	}
	pos = 0;
	for (i = 0; i < size; i++) {
		if (i % CHAR_BIT == 0)
        	flags = buf[pos++];
        else
        	flags <<= 1;
		if (flags & (1U << (CHAR_BIT - 1))) {
			encode_c(buf[pos++] + (1U << CHAR_BIT));
			k = buf[pos++] << CHAR_BIT;
            k += buf[pos++];
			encode_p(k);
		} else
        	encode_c(buf[pos++]);
	}
	for (i = 0; i < NC; i++)
    	c_freq[i] = 0;
	for (i = 0; i < NP; i++)
    	p_freq[i] = 0;
}


/* make tree, calculate len[], return root */
static int make_tree(int nparm, ushort freqparm[],
				uchar lenparm[], ushort codeparm[])
{
	int i, j, k, avail;

	n = nparm;
    freq = freqparm;
    len = lenparm;
	avail = n;
    heapsize = 0;
    heap[1] = 0;
	for (i = 0; i < n; i++) {
		len[i] = 0;
		if (freq[i])
        	heap[++heapsize] = i;
	}
	if (heapsize < 2) {
		codeparm[heap[1]] = 0;
        return heap[1];
	}
	for (i = heapsize / 2; i >= 1; i--)
		downheap(i);  /* make priority queue */
	sortptr = codeparm;
	do {  /* while queue has at least two entries */
		i = heap[1];  /* take out least-freq entry */
		if (i < n)
        	*sortptr++ = i;
		heap[1] = heap[heapsize--];
		downheap(1);
		j = heap[1];  /* next least-freq entry */
		if (j < n)
        	*sortptr++ = j;
		k = avail++;  /* generate new node */
		freq[k] = freq[i] + freq[j];
		heap[1] = k;
        downheap(1);  /* put into queue */
		left[k] = i;
        right[k] = j;
	} while (heapsize > 1);
	sortptr = codeparm;
	make_len(k);
	make_code(nparm, lenparm, codeparm);
	return k;  /* return root */
}


static void count_t_freq(void)
{
	int i, k, n, count;

	for (i = 0; i < NT; i++)
    	t_freq[i] = 0;
	n = NC;
	while (n > 0 && c_len[n - 1] == 0)
    	n--;
	i = 0;
	while (i < n) {
		k = c_len[i++];
		if (k == 0) {
			count = 1;
			while (i < n && c_len[i] == 0) {
            	i++;
                count++;
            }
			if (count <= 2)
            	t_freq[0] += count;
			else if (count <= 18)
            	t_freq[1]++;
			else if (count == 19) {
            	t_freq[0]++;
                t_freq[1]++;
            } else
            	t_freq[2]++;
		} else
        	t_freq[k + 2]++;
	}
}


static void write_pt_len(int n, int nbit, int i_special)
{
	int i, k;

	while (n > 0 && pt_len[n - 1] == 0)
    	n--;
	putbits(nbit, n);
	i = 0;
	while (i < n) {
		k = pt_len[i++];
		if (k <= 6)
        	putbits(3, k);
		else
        	putbits(k - 3, (1U << (k - 3)) - 2);
		if (i == i_special) {
			while (i < 6 && pt_len[i] == 0)
            	i++;
			putbits(2, (i - 3) & 3);
		}
	}
}


static void write_c_len(void)
{
	int i, k, n, count;

	n = NC;
	while (n > 0 && c_len[n - 1] == 0)
    	n--;
	putbits(CBIT, n);
	i = 0;
	while (i < n) {
		k = c_len[i++];
		if (k == 0) {
			count = 1;
			while (i < n && c_len[i] == 0) {
            	i++;
                count++;
            }
			if (count <= 2) {
				for (k = 0; k < count; k++)
					putbits(pt_len[0], pt_code[0]);
			} else if (count <= 18) {
				putbits(pt_len[1], pt_code[1]);
				putbits(4, count - 3);
			} else if (count == 19) {
				putbits(pt_len[0], pt_code[0]);
				putbits(pt_len[1], pt_code[1]);
				putbits(4, 15);
			} else {
				putbits(pt_len[2], pt_code[2]);
				putbits(CBIT, count - 20);
			}
		} else
        	putbits(pt_len[k + 2], pt_code[k + 2]);
	}
}


static void encode_c(int c)
{
	putbits(c_len[c], c_code[c]);
}


static void encode_p(uint p)
{
	uint c, q;

	c = 0;
    q = p;
    while (q) {
    	q >>= 1;
        c++;
    }
	putbits(pt_len[c], pt_code[c]);
	if (c > 1)
    	putbits(c - 1, p & (0xFFFFU >> (17 - c)));
}


/* priority queue; send i-th entry down heap */
static void downheap(int i)
{
	int j, k;

	k = heap[i];
	while ((j = 2 * i) <= heapsize) {
		if (j < heapsize && freq[heap[j]] > freq[heap[j + 1]])
		 	j++;
		if (freq[k] <= freq[heap[j]])
        	break;
		heap[i] = heap[j];
        i = j;
	}
	heap[i] = k;
}


static void make_len(int root)
{
	int i, k;
	uint cum;

	for (i = 0; i <= 16; i++)
    	len_cnt[i] = 0;
	count_len(root);
	cum = 0;
	for (i = 16; i > 0; i--)
		cum += len_cnt[i] << (16 - i);
	while (cum != (1U << 16)) {
		//fprintf(stderr, "17");
        error("17");
		len_cnt[16]--;
		for (i = 15; i > 0; i--) {
			if (len_cnt[i] != 0) {
				len_cnt[i]--;
                len_cnt[i+1] += 2;
                break;
			}
		}
		cum--;
	}
	for (i = 16; i > 0; i--) {
		k = len_cnt[i];
		while (--k >= 0)
        	len[*sortptr++] = i;
	}
}


static void make_code(int n, uchar len[], ushort code[])
{
	int    i;
	ushort start[18];

	start[1] = 0;
	for (i = 1; i <= 16; i++)
		start[i + 1] = (start[i] + len_cnt[i]) << 1;
	for (i = 0; i < n; i++)
    	code[i] = start[len[i]]++;
}


static void count_len(int i)  /* call with i = root */
{
	static int depth = 0;

	if (i < n)
    	len_cnt[(depth < 16) ? depth : 16]++;
	else {
		depth++;
		count_len(left [i]);
		count_len(right[i]);
		depth--;
	}
}


/*--------------------------------------------------------------------------*/
/* variable number of output bits                                           */
/*--------------------------------------------------------------------------*/


static void init_putbits(void)
{
	bitcount = CHAR_BIT;
    subbitbuf = 0;
}


/* write rightmost n bits of x */
static void putbits(int n, uint x)
{
	if (n < bitcount) {
		subbitbuf |= x << (bitcount -= n);
	} else {
		write_compressed(subbitbuf | (x >> (n -= bitcount)));
		if (n < CHAR_BIT) {
			subbitbuf = x << (bitcount = CHAR_BIT - n);
		} else {
			write_compressed(x >> (n - CHAR_BIT));
			subbitbuf = x << (bitcount = 2 * CHAR_BIT - n);
		}
	}
}

/*--------------------------------------------------------------------------*/
/*  Copyright (C) Herbert Oppmann 2003. All Rights Reserved.                */
/*--------------------------------------------------------------------------*/

