/*--------------------------------------------------------------------------*/
/*  Copyright (C) Herbert Oppmann 2003. All Rights Reserved.                */
/*--------------------------------------------------------------------------*/
/*  Project:     MyBCP                                                      */
/*  Filename:    LZHUnpack.c                                                */
/*                                                                          */
/*  Description:                                                            */
/*                                                                          */
/*  LZH decompression implementation                                        */
/*--------------------------------------------------------------------------*/

/*--< define, include >-----------------------------------------------------*/

#include "LZHapi.h"
#include "LZHint.h"

#include <stdio.h>

/*--< variables >-----------------------------------------------------------*/

static ushort	bitbuf;
static ushort	c_table[CTABLESIZE];
static ushort	pt_table[256];
static ushort	blocksize;

/*--< function declarations >-----------------------------------------------*/

/* static Huffman encoding */
static void		decode_start(void);
static ushort	decode_c(void);
static ushort	decode_p(void);
static void		read_pt_len(int nn, int nbit, int i_special);
static void		read_c_len(void);
static void		make_table(int nchar, uchar *bitlen, int tablebits, ushort *table);
/* variable number of input bits */
static void		init_getbits(void);
static ushort	getbits(int n);
static void		fillbuf(int n);

/*--< functions >-----------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* Lempel-Ziv 77 sliding window                                             */
/*--------------------------------------------------------------------------*/


void decode(ulong origsize)
{
	short i;		// copy index in window
	short j;		// copy length
	short c;		// decoded char or len
	short r;		// write index in window
	uchar *text;	// sliding window
	ulong count;	// uncompressed output length (0 .. origsize)

	text = (uchar *)malloc(DDICSIZ);

	decode_start();	// init huffman
	count = 0;
	r = 0;

	while (count < origsize)
	{
		if ((c = decode_c()) <= UCHAR_MAX)
		{
			text[r] = (uchar) c;
			count++;
			if (++r >= DDICSIZ)
		 	{
				r = 0;
				write_uncompressed(text, DDICSIZ);
			}
		}
		else
		{
			j = c - (UCHAR_MAX + 1 - THRESHOLD);
			count += j;
			i = decode_p();
			if ((i = r - i - 1) < 0)
				i += DDICSIZ;
			if (r > i && r < DDICSIZ - MAXMATCH - 1)
			{
				while (--j >= 0)
					text[r++] = text[i++];
			}
			else
			{
				while (--j >= 0)
				{
					text[r] = text[i];
					if (++r >= DDICSIZ)
					{
						r = 0;
						write_uncompressed(text, DDICSIZ);
					}
					if (++i >= DDICSIZ)
						i = 0;
				}
			}
		}
	}
	if (r != 0)
		write_uncompressed(text, r);

	free((char *)text);
}


/*--------------------------------------------------------------------------*/
/* static Huffman encoding                                                  */
/*--------------------------------------------------------------------------*/


/* init Huffman */
static void decode_start(void)
{
	blocksize = 0;
	init_getbits();	// init variable number of input bits
}


/* decode a char or len */
static ushort decode_c(void)
{
	ushort j, mask;

	if (blocksize == 0)
	{
		blocksize = getbits(16);
		read_pt_len(NT, TBIT, 3);
		read_c_len();
		read_pt_len(NP, PBIT, -1);
	}
	blocksize--;
	j = c_table[bitbuf >> 4];
	if (j >= NC)
	{
		mask = 1 << (3);
		do
		{
			if (bitbuf & mask)
				j = right[j];
			else
				j = left[j];
			mask >>= 1;
		} while (j >= NC);
	}
	fillbuf((int)(c_len[j]));
	return j;
}


/* decode a pointer */
static ushort decode_p(void)
{
	ushort j, mask;

	j = pt_table[bitbuf >> (8)];
	if (j >= NP)
	{
		mask = 1 << (7);
		do
		{
			if (bitbuf & mask)
				j = right[j];
			else
				j = left[j];
			mask >>= 1;
		} while (j >= NP);
	}
	fillbuf((int)(pt_len[j]));
	if (j != 0)
	{
		j--;
		j = (1 << j) + getbits((int)j);
	}
	return j;
}


static void read_c_len(void)
{
	short i, c, n;
	ushort mask;

	n = getbits(CBIT);
	if (n == 0)
	{
		c = getbits(CBIT);
		for (i = 0; i < NC; i++)
			c_len[i] = 0;
		for (i = 0; i < CTABLESIZE; i++)
			c_table[i] = c;
	}
	else
	{
		i = 0;
		while (i < n)
		{
			c = pt_table[bitbuf >> (8)];
			if (c >= NT)
			{
				mask = 1 << (7);
				do
				{
					if (bitbuf & mask)
						c = right[c];
					else
						c = left[c];
					mask >>= 1;
				} while (c >= NT);
			}
			fillbuf((int)(pt_len[c]));
			if (c <= 2)
			{
				if (c == 0)
					c = 1;
				else if (c == 1)
					c = getbits(4) + 3;
				else
					c = getbits(CBIT) + 20;
				while (--c >= 0)
					c_len[i++] = 0;
			}
			else
				c_len[i++] = (uchar)(c - 2);
		}
		while (i < NC)
			c_len[i++] = 0;
		make_table(NC, c_len, 12, c_table);
	}
}


static void read_pt_len(int nn, int nbit, int i_special)
{
	int i, n;
	short c;
	ushort mask;

	n = getbits(nbit);
	if (n == 0)
	{
		c = getbits(nbit);
		for (i = 0; i < nn; i++)
			pt_len[i] = 0;
		for (i = 0; i < 256; i++)
			pt_table[i] = c;
	}
	else
	{
		i = 0;
		while (i < n)
		{
			c = bitbuf >> (13);
			if (c == 7)
			{
				mask = 1 << (12);
				while (mask & bitbuf)
				{
					mask >>= 1;
					c++;
				}
			}
			fillbuf((c < 7) ? 3 : (int)(c - 3));
			pt_len[i++] = (uchar)c;
			if (i == i_special)
			{
				c = getbits(2);
				while (--c >= 0)
					pt_len[i++] = 0;
			}
		}
		while (i < nn)
			pt_len[i++] = 0;
		make_table(nn, pt_len, 8, pt_table);
	}
}


static void make_table(int nchar, uchar *bitlen, int tablebits, ushort *table)
{
	ushort count[17], weight[17], start[18], *p;
	uint i, k, len, ch, jutbits, avail, nextcode, mask;

	for (i = 1; i <= 16; i++)
		count[i] = 0;
	for (i = 0; (int)i < nchar; i++)
		count[bitlen[i]]++;

	start[1] = 0;
	for (i = 1; i <= 16; i++)
		start[i + 1] = start[i] + (count[i] << (16 - i));
		if (start[17] != (ushort) (1 << 16)) {
			printf("Bad table\n");
			exit(1);
		}

	jutbits = 16 - tablebits;
	for (i = 1; (int)i <= tablebits; i++)
	{
		start[i] >>= jutbits;
		weight[i] = 1 << (tablebits - i);
	}
	while (i <= 16)
	{
		weight[i] = 1 << (16 - i);
		i++;
	}

	i = start[tablebits + 1] >> jutbits;
	if (i != (ushort) (1 << 16))
	{
		k = 1 << tablebits;
		while (i != k)
			table[i++] = 0;
	}

	avail = nchar;
	mask = 1 << (15 - tablebits);
	for (ch = 0; (int)ch < nchar; ch++)
	{
		if ((len = bitlen[ch]) == 0)
			continue;
		k = start[len];
		nextcode = k + weight[len];
		if ((int)len <= tablebits)
		{
			for (i = start[len]; i < nextcode; i++)
				table[i] = ch;
		}
		else
		{
			p = &table[k >> jutbits];
			i = len - tablebits;
			while (i != 0)
			{
				if (*p == 0)
				{
					right[avail] = left[avail] = 0;
					*p = avail++;
				}
				if (k & mask)
					p = &right[*p];
				else
					p = &left[*p];
				k <<= 1;
				i--;
			}
			*p = ch;
		}
		start[len] = nextcode;
	}
}


/*--------------------------------------------------------------------------*/
/* variable number of input bits                                            */
/*--------------------------------------------------------------------------*/


/* init bitwise input */
static void init_getbits(void)
{
	bitbuf = 0;
	subbitbuf = 0;
	bitcount = 0;
	fillbuf(2 * CHAR_BIT);
}


/* get variable number of input bits */
static ushort getbits(int n)
{
	ushort x;

	x = bitbuf >> (2 * CHAR_BIT - n);
	fillbuf(n);
	return x;
}


/* shift bitbuf n bits left, read n bits */
static void fillbuf(int n)
{
	bitbuf = (bitbuf << n) & 0xFFFF;  /* lose the first n bits */
	while (n > bitcount) {
		bitbuf |= subbitbuf << (n -= bitcount);

		subbitbuf = read_compressed();
		bitcount = CHAR_BIT;
	}
	bitbuf |= subbitbuf >> (bitcount -= n);
}

/*--------------------------------------------------------------------------*/
/*  Copyright (C) Herbert Oppmann 2003. All Rights Reserved.                */
/*--------------------------------------------------------------------------*/

